#!/usr/lib/perl5 -w

####################################################################################
## cliqr2.pl
## #################
## Script used by the cliqr GUI to search the extended best track database of ATL
## systems back to 1851 (extended part only goes back to 1980 at best, and has gaps)
## Points awarded based on various measurements of the storm, and these points are
## all from initial conditions vs. a single point's measurements. Future plans include
## possibly including the track forecast, but this will require a large change in
## implementation.
####################################################################################

$GRPHGD=$ENV{'GRPHGD'};

## Needed math functions!! for distance/direction/speed calculations
use Math::Trig qw(great_circle_distance deg2rad rad2deg great_circle_direction);
use Math::Complex qw(Re);

## Open the massive database file
open(DAT,"$GRPHGD/cliqr_new/ebtrk_nhc_final.txt")||die("File cannot be opened");
@data=<DAT>;
my ($i,$points,$pointsawarded,$hn)=0;

## initial parameters /from CHGHUR/
my ($ilat,$ilon,$icenpres,$ioutpres,$ioutrad,$imxwnd,$idir,$ispd)=0;

## limit parameters, mostly based on initial geographic location
my ($latl,$lonl,$cenpresl,$ouptresl,$outradl,$mxwndl,$dirl,$spdl)=0;

## matching arrays
my (@points,@pointmatch,@pointmatch2,@finallist,@ptprev,@ptcur) = {};

## Setting defaults for parameter limits

$latl=3;
$lonl=3;
$cenpresl=8;
$outpresl=3;
$outradl=30;
$mxwndl=10;
$dirl=15;

## Reading data into line-by-line array from ebtrk file
## As of 7/15/08, file was ~46400 lines.
## As of 5/27/08, file was ~67000 lines.

foreach $line(@data) 
{
  chomp $line;
  $points[$i]=$line;
  $i++;
}
close DAT;
my $numpoints=$i-1;
$i=0;

## Setting intial data from arguments passed to script

$ilat=$ARGV[0];
$ilon=$ARGV[1];
$icenpres=$ARGV[2];
$ioutpres=$ARGV[3];
$ioutrad=$ARGV[4];
$imxwnd=$ARGV[5];
$idir=$ARGV[6];
$ispd=$ARGV[7];

print $ilat,$ilon,$icenpres,$ioutpres,$ioutrad,$imxwnd,$idir,$ispd;
print "\n";

foreach (@ARGV)
{
  print $_."\n";
}

###############################################################################################################
## Matching the inital point of the system to others that had any recorded points within the location-defined
## lat and lon range.
## 'sub locs' is used for converting lat/lon into radians for the distance/direction calculations
## This section currently performs far more calculations than necessary, as distance is computed for all lines
## instead of just those being considered after the lat/lon seach.
###############################################################################################################

sub locs { deg2rad($_[0]), deg2rad(90 - $_[1]) }
for($i=0;$i<@data;$i++)
{
  ## Getting two lat/lons to compare

  $line=$data[$i];
  $lat=substr($line,29,4);
  $plat=substr($data[$i-1],29,4);
  $lon=substr($line,34,5);
  $plon=substr($data[$i-1],34,5);

  ## Converting to radians for calculations

  @ptprev = locs($plon,$plat);
  @ptcur = locs($lon, $lat);

  ## If the potential match is within the lat/lon limits,
  ## make it a match!
  if($lat>($ilat-$latl)&&$lat<($ilat+$latl))
  {
    if($lon>($ilon-$lonl)&&$lon<($ilon+$lonl))
    {
      push(@pointmatch,$line);
    }
  }

  ## Sorting by geographic boxes

  if($lat<=20&&$lon<=50)
  {
    $outpresl=2;
  }
  if($lat<25&&$lon>50&&$lon<=75)
  {
    $latl=2;
    $lonl=2;
    $outpresl=2;
    $mxwndl=5;
    $cenpresl=6;
  }
  if($lat<32&&$lon>75)
  {
    $latl=2;
    $lonl=2;
    $cenpresl=5;
    $outpresl=2;
    $mxwndl=5;
    $outradl=20;
  }
  if($lat>=32&&$lat<=50&&$lon>=70)
  {
    $latl=2;
    $lonl=2;
    $cenpresl=6;
    $outradl=20;
  }
  if(($lat>20&&$lon<=50)||$lat>50)
  {
    $outradl=40;
  }
}

for($i=0;$i<@pointmatch;$i++)
{
  ## Getting two lat/lons to compare

  $line=$pointmatch[$i];
  $lat=substr($line,29,4);
  $plat=substr($pointmatch[$i-1],29,4);
  $lon=substr($line,34,5);
  $plon=substr($pointmatch[$i-1],34,5);

  ## Converting to radians for calculations

  @ptprev = locs($plon,$plat);
  @ptcur = locs($lon, $lat);
  ## Compute distance, convert from kms to miles

  if (($plon==$lon) && ($plat==$lat))
  {
    $dist=0;
  }else{
    $dist = great_circle_distance(@ptprev, @ptcur, 6378);
    $dist = int($dist/1.852);
  }

  #############################################################################################################
  ## NOTE: This next section is intended to eliminate and readjust cases where distance is calculated from the
  ## last point of one storm to the first point of the next. Currently, all it does is assume that if the
  ## calculated distance is greater than 600 miles that it must be two seperate storms and if it is less than
  ## 600 miles than it must be the same storm. However, this is a seriously flawed system and may result in
  ## inaccurate results when two consecutive storms are in the Gulf of Mexico or when one storm dies and the
  ## subsequent storm in the database forms fairly close to the last position of the previous storm. I have
  ## also noted issues in the EPAC, with many of the same problems as listed before but also other potential
  ## issues, such as distance/direction group on the first point of Howard 2004 is '204 143' which would appear
  ## to be closest to the direction from the previous point of Georgette, but that is well over 600 miles 
  ## away. Lines included below for reference.
  ## EP1004 GEORGETTE 090312 2004 23.1 146.1  15 1013 -99 -99  -99 -99   0  0  0  0   0  0  0  0   0  0  0  0 L
  ## EP1104 HOWARD    083012 2004 11.9 103.1  25 1008 -99 -99 1009 150   0  0  0  0   0  0  0  0   0  0  0  0 *
  #############################################################################################################

  if ($dist>600)
  {
    $plat=substr($pointmatch[$i+1],29,4);
    $plon=substr($pointmatch[$i+1],34,5);
    @ptprev = locs($plon,$plat);
    $dist = great_circle_distance(@ptprev, @ptcur, 6378);
    $dist = int($dist*1.852);
  }

  ## Calculate direction now that distance has been corrected
 
  ##print "-- $plat $plon () $lat $lon --\n"; 
  if ($plon==$lon)
  {
    if ($plat==$lat)
    {
      $dir=0;
    } else {
      if($plat>$lat)
      {
        $dir=180
      } else {
        $dir=0;
      }
    }
  } else {
    ##print ">>inside<<\n";
    $dir = Re(great_circle_direction(@ptprev, @ptcur)); 
    ##print "dir1=$dir\n";
    $dir = int(rad2deg($dir));
    ##print "dir2=$dir\n\n";
  }

  ## Fix direction if it is sub-zero or over 360

  if($dir<0) { $dir = $dir +360; }

  ## Formatting and appending on to end of line

  for($j=1000;$j>$dist&&$j>1;$j=$j/10)
  {
    $line=$line." ";
  }
  $line=$line.$dist;
  for($j=1000;$j>$dir&&$j>1;$j=$j/10)
  {
    $line=$line." ";
  }
  $line=$line.$dir;
  $line=$line."  0.0";

  push(@pointmatch2,$line);
}

## Loop for awarding points to each line one-by-one and appending final score at end of line

for($i=0;$i<@pointmatch2;$i++)
{
  $pointsawarded=0;

  ## Points for direction: 2pts if within bounds, 0.8pts if within bounds+1/3, 0.4pts if within bounds+1/2

  $dir=substr($pointmatch2[$i],111,3);
  $idirmin=$idir-$dirl;
  $idirmax=$idir+$dirl;
  if($idirmin<0) {$idirmin=$idirmin+360;}
  if($idirmax>=360) {$idirmax=$idirmax-360;}
  if($idir<(360-$dirl)&&$idir>($dirl))
  {
    if($dir>($idirmin)&&$dir<($idirmax))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+2.0;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
  }
  else {
    if($dir>($idirmin)||$dir<($idirmax))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+2.0;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
  }
  if($pointsawarded==0)
  {
    $idirmin=$idir-($dirl*(4/3));
    $idirmax=$idir+($dirl*(4/3));
    if($idirmin<0) {$idirmin=$idirmin+360;}
    if($idirmax>=360) {$idirmax=$idirmax-360;}
    if($idir<(360-$dirl)&&$idir>($dirl))
    {
      if($dir>($idirmin)&&$dir<($idirmax))
      {
        $points=substr($pointmatch2[$i],115,4);
        $points=$points+0.8;
        $pointsawarded=1;
        if(int($points)==$points)
        {
          $points=$points.".0";
        }
        $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
      }
    }
    else {
      if($dir>($idirmin)||$dir<($idirmax))
      {
        $points=substr($pointmatch2[$i],115,4);
        $points=$points+0.8;
        $pointsawarded=1;
        if(int($points)==$points)
        {
          $points=$points.".0";
        }
        $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
      }
    }
    if($pointsawarded==0)
    {
      $idirmin=$idir-($dirl*(3/2));
      $idirmax=$idir+($dirl*(3/2));
      if($idirmin<0) {$idirmin=$idirmin+360;}
      if($idirmax>=360) {$idirmax=$idirmax-360;}
      if($idir<(360-$dirl)&&$idir>($dirl))
      {
        if($dir>($idirmin)&&$dir<($idirmax))
        {
          $points=substr($pointmatch2[$i],115,4);
          $points=$points+0.4;
          $pointsawarded=1;
          if(int($points)==$points)
          {
            $points=$points.".0";
          }
          $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
        }
      }
      else {
        if($dir>($idirmin)||$dir<($idirmax))
        {
          $points=substr($pointmatch2[$i],115,4);
          $points=$points+0.4;
          $pointsawarded=1;
          if(int($points)==$points)
          {
            $points=$points.".0";
          }
          $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
        }
      }
    } 
  }
  $pointsawarded=0;

  ## Points awarded for ROCI: 3pts if within bounds, 1.5pts if within bounds+1/3, 0.7pts if within bounds+1/2

  $outrad=substr($pointmatch2[$i],62,3);
  if($outrad>($ioutrad-$outradl)&&$outrad<($ioutrad+$outradl))
  {
    $points=substr($pointmatch2[$i],115,4);
    $points=$points+3.0;
    $pointsawarded=1;
    if(int($points)==$points)
    {
      $points=$points.".0";
    }
    $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
  }
  if($pointsawarded==0)
  {
    if($outrad>($ioutrad-($outradl*(4/3)))&&$outrad<($ioutrad+($outradl*(4/3))))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+1.5;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
    if($pointsawarded==0)
    {
      if($outrad>($ioutrad-($outradl*(3/2)))&&$outrad<($ioutrad+($outradl*(3/2))))
      {
        $points=substr($pointmatch2[$i],115,4);
        $points=$points+0.7;
        $pointsawarded=1;
        if(int($points)==$points)
        {
          $points=$points.".0";
        }
        $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
      }
    }
  }
  $pointsawarded=0;

  ## Points awarded for central pressure: 1pt if within bounds, 0.5pt if within bounds+1/4, 0.2pts if within bounds+1/3

  $cenpres=substr($pointmatch2[$i],44,4);
  if($cenpres>($icenpres-$cenpresl)&&$cenpres<($icenpres+$cenpresl))
  {
    $points=substr($pointmatch2[$i],115,4);
    $points=$points+1.0;
    $pointsawarded=1;
    if(int($points)==$points)
    {
      $points=$points.".0";
    }
    $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
  }
  if($pointsawarded==0)
  {
    if($cenpres>($icenpres-($cenpresl*(5/4)))&&$cenpres<($icenpres+($cenpresl*(5/4))))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+0.5;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
    if($pointsawarded==0)
    {
      if($cenpres>($icenpres-($cenpresl*(4/3)))&&$cenpres<($icenpres+($cenpresl*(4/3))))
      {
        $points=substr($pointmatch2[$i],115,4);
        $points=$points+0.2;
        $pointsawarded=1;
        if(int($points)==$points)
        {
          $points=$points.".0";
        }
        $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
      }
    }
  }
  $pointsawarded=0;

  ## Points awarded for outer pressure: 0.5pts if within bounds, 0.2pts if within bounds+1/2

  $outpres=substr($pointmatch2[$i],57,4);
  if($outpres>($ioutpres-$outpresl)&&$outpres<($ioutpres+$outpresl))
  {
    $points=substr($pointmatch2[$i],115,4);
    $points=$points+0.5;
    $pointsawarded=1;
    if(int($points)==$points)
    {
      $points=$points.".0";
    }
    $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
  }
  if($pointsawarded==0)
  {
    if($outpres>($ioutpres-($outpresl*(3/2)))&&$outpres<($ioutpres+($outpresl*(3/2))))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+0.2;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
  }
  $pointsawarded=0;

  ## Points awarded for max wind: 0.8pts if within bounds, 0.3pts if within bounds+5kts

  $mxwnd=substr($pointmatch2[$i],40,3);
  if($mxwnd>=($imxwnd-$mxwndl)&&$mxwnd<=($imxwnd+$mxwndl))
  {
    $points=substr($pointmatch2[$i],115,4);
    $points=$points+0.8;
    $pointsawarded=1;
    if(int($points)==$points)
    {
      $points=$points.".0";
    }
    $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
  }
  if($pointsawarded==0)
  {
    if($mxwnd>=($imxwnd-($mxwndl-5))&&$mxwnd<=($imxwnd+($mxwndl+5)))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+0.3;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
  }
  $pointsawarded=0;

  ## Points awarded for speed: 2pts if within bounds, 1pt if within bounds+1/4, 0.5pts if within bounds+1/3

  $dist=substr($pointmatch2[$i],107,3);
  $spd=$dist/6;
  if($spd<1.5)
  {
    $spd=$spd+1
  }
  $spdl=$spd*.2;
  if($spd>=($ispd-$spdl-.5)&&$spd<=($ispd+$spdl+.5))
  {
    $points=substr($pointmatch2[$i],115,4);
    $points=$points+2.0;
    $pointsawarded=1;
    if(int($points)==$points)
    {
      $points=$points.".0";
    }
    $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
  }
  if($pointsawarded==0)
  {
    if($spd>=($ispd-($spdl*(5/4))-.5)&&$spd<=($ispd+($spdl*(5/4)).5))
    {
      $points=substr($pointmatch2[$i],115,4);
      $points=$points+1.0;
      $pointsawarded=1;
      if(int($points)==$points)
      {
        $points=$points.".0";
      }
      $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
    }
    if($pointsawarded==0)
    {
      if($spd>=($ispd-($spdl*(4/3))-.5)&&$spd<=($ispd+($spdl*(4/3)).5))
      {
        $points=substr($pointmatch2[$i],115,4);
        $points=$points+0.5;
        $pointsawarded=1;
        if(int($points)==$points)
        {
          $points=$points.".0";
        }
        $pointmatch2[$i]=substr($pointmatch2[$i],0,115)." ".$points;
      }
    }
  }
}

## The awkward way of sorting by final score, for some reason doesn't like
## to incrememnt by .1 in the for loop, so increment by 1 and divide by 10

for($i=110;$i>=0;$i--)
{
  foreach(@pointmatch2)
  {
    if(substr($_,115,4)==($i/10))
    {
      push(@finallist,$_);
    }
  }
}

## Print list statements

chdir "cliqrwork/cliqr";
open(PUT,">cliqrout.txt");

foreach (@finallist)
{
  if(substr($_,115,4)!=0.0)
  {
    print PUT $_."\n";
  }
}
print "Successfully completed search.";

## If of zero size, put "no reults found" in file.
## Sometimes happens on small/short files as well
## due to perl's lack of pausing long enough to write
## something before moving on to the command below.
## NOTE: This should be corrected by this sleep 1 below.

system "sleep 1";

if(-z "cliqrout.txt")
{
  print PUT "No results found.";
}
